##  columnctrl.tcl
##
##	Column-Control : add-on for tablelist widget.
##      Add a button for interactively hide/show tablelist columns,
##       and for change the columns order.
##
##  Copyright (c) 2005 <Irrational Numbers> : <aldo.buratti@tiscali.it> 
##
##  NOTE: tablelist package is freely downloadable from the Web:
##    http://www.nemethi.de
##
##  This program is free software; you can redistibute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation.
##
##  See the file "license" for information on usage and
##  redistribution of this program, and for a disclaimer of all warranties.
##


namespace eval tablelist::columnctrl {
    namespace export attach detach controlButton controlMenu configure
    
    # max number of popup's elements displayed
     # may be set externally (with care)
    variable popupLimit    10
    
    variable columnTrace; # array
    variable image;       #array
    variable widget;      #array
    
     # ? should you use classic widgets or tile-widgets ?
     # it depends on tablelist::usingTile ...
    if { [info exists tablelist::usingTile] && $tablelist::usingTile } {
        set widget(button)    ttk::button
        set widget(checkbutton)    ttk::checkbutton
        set widget(frame)     ttk::frame
        set widget(label)     ttk::label
        package require ttk::scrodget 2.1
        set widget(scrodget)  ttk::scrodget
    } else {
        set widget(button)    button
        set widget(checkbutton)    checkbutton
        set widget(frame)     frame
        set widget(label)     label
        package require scrodget 2.1
        set widget(scrodget)  scrodget
    }


    # create a button, and a menu, both children of $tbl 
    #  (so that when $tbl is destroyed, they are destroyed, too)
    proc attach {tbl} {
        variable DATA
        variable image
        variable widget

        $widget(button) [controlButton $tbl] -image $image(startbutton) \
                -command [list [namespace current]::_popup $tbl]
        place [controlButton $tbl] -in [$tbl bodypath] -relx 1.0 -anchor ne
        set DATA($tbl,dummy) 1
        set DATA($tbl,-nevershow) {}
        set DATA($tbl,-neverhide) {}
         # create a (dummy) popup menu
        menu [controlMenu $tbl] -tearoff 0
         # when controlMenu is destroyed
         #  - directly by "detach"
         #  - indirectly by destroying $tbl
         # remove all associated data
        bind [controlMenu $tbl] <Destroy> [list [namespace current]::_onDestroy $tbl]
    }
    
    proc detach {tbl} {
        destroy [controlButton $tbl]
        destroy [controlMenu $tbl]
    }
    
     # to do: configure should ALSO work in query mode;
     #  i.e. whane called without parameters, it should list all options/values.
     #       when called with one parameter, it should list that option/value
     
     # NOTE:
     #  configuring a 'not-yet-attached' columnctrl, has no effect.
     #  Configure only works on 'attached' columnctrl.
    proc configure {tbl args} {
        variable DATA

        if { ! [info exists DATA($tbl,dummy)] } return
        array set options $args
        foreach {opt val} [array get options] {
            switch -- $opt {
              -neverhide -
              -nevershow { set DATA($tbl,$opt) $val }
              default { error "Unknown option \"$opt\"" }
            }
        }
    }
    
     # returns the pathname of the 'control' button 
     # use with care (just for setting look&feel ..)
    proc controlButton {tbl} {
        return $tbl.columnCtrl
    }

     # returns the pathname of the column-menu attached to $tbl  
     # use with care (just for setting look&feel ..)
    proc controlMenu {tbl} {
        return $tbl.columnMenu
    }
}




proc tablelist::columnctrl::_init {} {
    variable image;
    
    # load the message catalog
    set msgdir [file join [file dirname [info script]] msgs]
    set res [::msgcat::mcload $msgdir]
    if { $res == 0 } {
        set lang [::msgcat::mclocale]
        source [file join $msgdir default.msg]
    }    
    #
    # Create images
    #
    set imgdir [file join [file dirname [info script]] images]
    set image(checked)    [image create photo -file [file join $imgdir checked.gif]]
    set image(unchecked)  [image create photo -file [file join $imgdir unchecked.gif]]
    set image(startbutton) [image create photo -file [file join $imgdir col3small.gif]]
}


proc tablelist::columnctrl::_onDestroy {tbl} {
    variable DATA
    array unset DATA $tbl,*
}


proc tablelist::columnctrl::_getColumns {tbl {limit {}}} {
    variable DATA
    
    # could you imagine a table with one million columns ??
    if { $limit == {} } { set limit 1000000 }
    set L {}
    
    set maxC [$tbl columncount]
    # number of column found (excluding always-hidden columns)
    set nC  0
    for { set i 0 } { $i < $maxC } { incr i } {
        if { $nC == $limit } break
        
        set colName [$tbl columncget $i -name]
        if { [lsearch $DATA($tbl,-nevershow) $colName] != -1 } continue
        
        lappend L [$tbl columncget $i -name] \
                [$tbl columncget $i -title] \
                [expr ! [$tbl columncget $i -hide]]
        incr nC
    }
    return $L
}


    # L is a list made of triplets
    #  { id label boolean   .... }
proc tablelist::columnctrl::_popup { tbl } {
    variable DATA

    # a global array is required for tracing input changes.
    # Since this proc is synchronous
    #  (i.e. there cannot be two menus at the same time),
    #  it is OK using a single global array
    
    variable columnTrace
    variable popupLimit
    
    set x [winfo pointerx $tbl]
    set y [winfo pointery $tbl]
    
    set m [controlMenu $tbl]
    $m delete 0 end
    array unset columnTrace
    
    set row 0
    foreach { id lbl val } [_getColumns $tbl $popupLimit]  {
        set columnTrace($id) $val
        if { $lbl == {} } { set lbl $id }
        $m add checkbutton -label $lbl -variable [namespace current]::columnTrace($id) \
           -command [list [namespace current]::_onChanged $tbl $id]
        if { [lsearch $DATA($tbl,-neverhide) $id] != -1 } {
            $m entryconfigure end -state disabled
        }
        incr row
    }
    $m add separator
    $m add command -label [::msgcat::mc more_columns] -command [list [namespace current]::_detailWindow $tbl]
    
    $m post $x $y
}


proc tablelist::columnctrl::_onChanged { tbl idx } {
    variable columnTrace
    set val [set columnTrace($idx)]
    $tbl columnconfigure $idx -hide [expr ! $val]
}


proc tablelist::columnctrl::_emptyStr   val { return "" }

proc tablelist::columnctrl::_editEndCmd {tbl row col txt} {
    variable image
    
    switch -- [$tbl columncget $col -name] {
        visible {
            #
            # Update the image contained in the cell
            #
            set img [expr {$txt ? $image(checked) : $image(unchecked)}]
            $tbl cellconfigure $row,$col -image $img
        }
    }
    return $txt
}



proc tablelist::columnctrl::_detailWindow { tbl } {
    variable DATA    
    variable widget

    set top [toplevel [controlButton $tbl].detail]
    wm title $top [::msgcat::mc columns_setting_title]
    
     #
     # If you are using tile, you need to create a frame covering the toplevel,
     #  or you will see unpleasant background around labels, etc..
     #
    set w [$widget(frame) $top.f]
    if { [winfo class $w] == "TFrame" } {
        $w configure -padding 10
    } else {
        $w configure -padx 10 -pady 10
    }
    pack $w -expand 1 -fill both

    $widget(label) $w.l1 -text [::msgcat::mc help_reorder]
    $widget(label) $w.l2 -text [::msgcat::mc help_hide]
    
    ::tablelist::tablelist $w.cols \
            -movablecolumns 0 \
            -selectmode single \
            -columns {
                0 "ID"      left
                0 ""	  center
                0 "Name"		  left
            }  \
            -activestyle none \
            -showlabels 0 \
            -editendcommand [namespace current]::_editEndCmd
    if { [$tbl cget -movablecolumns] } {
        $w.cols configure -movablerows 1
    } else {
        $w.cols configure -movablerows 0
        $w.l1 configure -state disabled
    }
    
    $w.cols columnconfigure 0 -name ID -hide 1
    
     # do not use ttk::checkbutton,
     #  because its look is different from the static images
    $w.cols columnconfigure 1 -name visible -editable yes \
            -editwindow checkbutton \
            -formatcommand [namespace current]::_emptyStr
    
    set row 0
    foreach { id lbl val } [_getColumns $tbl]  {
        if { $lbl == {} } { set lbl $id }
        $w.cols insert end [list $id $val $lbl]
        if { [lsearch $DATA($tbl,-neverhide) $id] != -1 } {
            $w.cols rowconfigure end  -selectable false \
               -foreground [$w.cols cget -disabledforeground]
            $w.cols cellconfigure end,visible -editable false
        }
        _editEndCmd $w.cols $row 1 $val
        incr row
    }
    
    $widget(scrodget) $w.sc -scrollsides w -autohide 1
    $w.sc associate $w.cols
    
    
    $widget(frame) $w.butt
    $widget(button) $w.butt.b1 -text [::msgcat::mc ok] -width 9 \
            -command "[namespace current]::_applyNewSettings $tbl $w.cols; destroy $top"
    $widget(button) $w.butt.b2 -text [::msgcat::mc cancel] -width 9 \
                    -command "destroy $top"

     # NOTE: packing order is important:
     #      buttons at bottom first !
    pack $w.butt.b2 $w.butt.b1 -side right -padx 5 -pady 5 
    pack $w.butt -side bottom -fill x
    
    pack $w.l1 $w.l2
    pack $w.sc -expand 1 -fill both -padx 5 -pady 5

    Dialog::setModalDialog $top
     # center the window
    ::tk::PlaceWindow $top
}

proc tablelist::columnctrl::_applyNewSettings {tbl t1} {
    $t1 finishediting
    set pos 0
    foreach c [$t1 get 0 end] {
        foreach { id val lbl } $c {}
        catch { $tbl movecolumn $id $pos }
        incr pos
        $tbl columnconfigure $id -hide [expr ! $val]
    }
}

tablelist::columnctrl::_init


package provide Tablelist::ColumnCtrl 1.1.1

